/* ascanner.c
**
** scan header elements
**
*/

#include "tools.h"
#include "vxerror.h"
#include "inline-funcs.h"
#include "mrmacs.h"

#include "rfc822_address.h"

/* Das ist die (in Standard-EBNF bertragene) Syntax fr E-Mail-Adressen
** gem RFC 822.
**
** address	: mailbox | group .
** group	: [ phrase ] ':' [ { mailbox ',' } mailbox ] ';' .
** mailbox	: addr-spec | phrase route-addr .
** route-addr	: '<' [ { '@' domain ',' } '@' domain ':' ] addr-spec '>' .
** addr-spec	: user [ '@' domain ] .
** user		: word { '.' word } .
** domain	: sub-domain { '.' sub-domain } .
** sub-domain	: atom | '[' literal ']' .
** atom		: atom-char { atom-char } .
** atom-char	: [^()<>@,;:\\".\[\] \0-\0x1f\0x7f] .
** lit-char	: [^\[\]\\\0x0d] | '\' <char> .
** literal	: lit-char { lit-char } .
** phrase	: word { word } .
** word-char	: [^"\\\0x1d] | '\' <char> .
** word		: atom | '"' { word-char } '"' .
*/
/* Hmmm,
** Die Syntax fr E-Mail-Adressen ist offensichtlich nicht ganz einfach.
** Offensichtlich mu ich zunchst von einer `phrase' ausgehen; wenn diese
** einem `user' entspricht und direkt im Anschlu ein `@' folgt, dann
** ist das eine Mailbox-Adresse.
** Folgt hingegen ein `<' oder ein `:', dann handelt es sich offensichtlich
** um eine Routing- oder eine Gruppenadresse.
** Folgt einem `word' einer Phrase kein `.', dann kann es sich ebenfalls nicht
** mehr um eine Mailbox-Adresse handeln.
*/

/* Die einzelnen Routinen zum Parsieren der E-Mail-Adressen entsprechen im
** wesentlichen den Produktionen in der oben angegebenen Syntax.
*/

/* p_atom parsiert ein `atom' aus der Zeichenkette, auf die der als erstes
** Argument bergebene Zeiger zeigt. Der Zeiger wird dabei direkt hinter
** das letzte Zeichen des gefundenen Atoms gesetzt. Ergebnis ist das gefundene
** Atom (als Zeichenkette). Das zweite Argument enthlt - im Fehlerfall -
** eine Meldung.
*/
static const char *noatom = "^[]<>@,;:\\\".() \0\1\2\3\4\5\6\7\10\11\12\13"
			    "\14\15\16\17\20\21\22\23\24\25\26\27\30\31\32"
			    "\33\34\35\36\37\177";
static char *p_atom (char **_p, char **error)
{
    char *p = *_p, *q = p, *res;
    while (*p && isNULL(strchr (noatom, *p))) { ++p; }
    if (p == q) { *_p = p; *error = "empty atom"; return NULL; }
    res = galloc (__FILE__, __LINE__, (p - q) + 1);
    memcpy (res, q, (p - q));
    res[p - q] = '\0';
    *_p = p;
    return res;
}

/* p_word parsiert ein `word', d.h. entweder ein Atom oder eine in `"'
** eingeschlossene Zeichenfolge. Die Argumente haben die gleiche
** Bedeutung wie bei `p_atom'.
*/
static char *p_word (char **_p, char **error)
{
    char *p = *_p, *q = p, *res;
    ifnull (strchr (noatom, *p)) { return p_atom (_p, error); }
    if (*p != '"') {
	*_p = p; *error = "non-atom word MUST be quoted"; return NULL;
    }
    ++p;
    while (*p && *p != '"') {
	if (*p == '\\') {
	    ++p;
	    if (*p == '\15' || *p == '\12') {
		*_p = p; *error = "multi-line word not allowed"; return NULL;
	    }
	}
	++p;
    }
    if (!*p) { *_p = p; *error = "missing closing quote"; return NULL; }
    ++p;
    res = galloc (__FILE__, __LINE__, (p - q) + 1);
    memcpy (res, q, (p - q));
    res[p - q] = '\0';
    *_p = p;
    return res;
}

/* p_literal parsiert ein `literal', d.h. entweder ein Atom oder eine in `['
** und `]' eingeschlossene Zeichenfolge. Die Argumente haben die gleiche
** Bedeutung wie bei `p_atom'.
*/
static char *p_literal (char **_p, char **error)
{
    char *p = *_p, *q = p, *res;
    ifnull (strchr (noatom, *p)) { return p_atom (_p, error); }
    if (*p != '[') {
	*_p = p;
	*error = "non-atom literal must be enclosed in `[' ans `]'";
	return NULL;
    }
    ++p;
    while (*p && *p != ']') {
	if (*p == '\\') {
	    ++p;
	    if (*p == '\15' || *p == '\12') {
		*_p = p;
		*error = "multi-line literal not allowed"; return NULL;
	    }
	} else if (*p == '[') {
	    *_p = p;
	    *error = "inside a literal, a `[' must be quoted"; return NULL;
	}
	++p;
    }
    if (!*p) { *_p = p; *error = "missing closing `]'"; return NULL; }
    res = galloc (__FILE__, __LINE__, (p - q) + 1);
    memcpy (res, q, (p - q));
    res[p - q] ='\0';
    *_p = p;
    return res;
}

/* p_comment parsiert einen `comment', d.h. eine in `[' und `]' eingeschlos-
** sene Zeichenfolge. Die Argumente haben die gleiche Bedeutung wie bei
** `p_atom'.
*/
static char *p_comment (char **_p, char **error)
{
    char *p = *_p, *r, *q = p, *res;
    int cl = 1;
    ++q; ++p;
    for (; *p; ++p) {
	if (*p == '(') { ++cl; continue; }
	if (*p == ')') { if (--cl == 0) break; continue; }
	if (*p == '\\') {
	    if (*++p == '\0') {
		*_p = p; *error = "invalid `\\' character"; return NULL;
	    }
	    continue;
	}
    }
    if (!*p) { *error = "missing closing `)'"; return NULL; }
    r = p; ++p; *_p = p;
    res = galloc (__FILE__, __LINE__, (r - q) + 1);
    memcpy (res, q, (r - q));
    res[r - q] = '\0';
    return res;
}

/* p_subdomain parsiert einen Abschnitt eines Domainnamens, d.h. entweder
** ein Atom oder ein (in `[' und `]' eingeschlossenes) Literal.
** Die Argumente haben die gleiche Bedeutung wie bei `p_atom'.
*/
static char *p_subdomain (char **_p, char **error)
{
    if (**_p == '[') { return p_literal (_p, error); }
    ifnull (strchr (noatom, **_p)) { return p_atom (_p, error); }
    *error = "invalid sub-domain";
    return NULL;
}

/* apndfree ist eine kleine Hilfsfunktion, um eine `phrase' bzw. mehrere
** Kommentare zu einem zusammenzufassen.
** Das erste Argument wird dabei mit dem zweiten verknpft und modifiziert
** zurckgegeben. Das zweite Argument wird direkt im Anschlu an die
** Verknpfung gelscht. (Daraus folgt, da das zweite Argument eine
** dynamisch erzeugte Zeichenkette sein mu.) Das dritte Argument sagt
** aus, ob zwischen die beiden ersten Argumente bei der Verknpfung ein
** Leerzeichen eingefgt werden soll (withbl != 0) oder nicht (withbl = 0).
*/
static __inline__ void apndfree (char **_out, char *in, int withbl)
{
    char *out;
    ifnull (*_out) { *_out = in; return; }
    ifnull (in) { return; }
    out = gconc (__FILE__, __LINE__, *_out, (withbl? " " : ""), in, NULL);
    gfree (*_out); gfree (in);
    *_out = out;
}

/* Da die Parsierung del `local part' und des `domain part' sehr hnlich
 * abluft, habe ich beides zusammengefat. Der wesentliche Unterschied
 * besteht nmlich nur in der Parsierung der einzelnen Abschnitte eines
 * `local part' (`username') oder `domain part'. Dazu wird die passende
 * Parsierungsfunktion als drittes Argument bergeben. Das erste Argument
 * hat die gleiche Bedeutung wie bei `p_atom', das zweite Argument ist ein
 * in/out-Argument, das zur Verkettung der evtl. vorkommenden Kommentare
 * verwendet wird. Das vierte Argument gibt jeweils die Zeichen an, die
 * das Ende eines `local part' bzw. eines `domain part' darstellen.
 * Das fnfte Argument
*/
static char *p_ud (char **_p, char **_comment, char *(*pf)(char **, char **error),
		   char *delim, int dodot, char **error)
{
    char *p = *_p;
    char *comment = NULL;
    char *res = NULL, *sd = NULL;
    /* The variable `dot' (a boolean switch) ensures that no sub-domain between
    ** two `.' can be empty; it also ensures that the domain name cannot begin
    ** or end with a `.'
    */
    int dot = 1;
    /* This macro should make the code of the function a bit mor readable.
    ** It ensures a consistent state before returning NULL in the case of
    ** an error.
    */
#   define FAILRET do { \
	*_p = p; \
	unlessnull (res) { gfree (res); res = NULL; } \
	unlessnull (comment) { gfree (comment); comment = NULL; } \
	return NULL; \
    } while (0)

    /* The incoming value of `_comment' MUST be copied in order to leave it
    ** in an unchanged state when returning an error.
    */
    unlessnull (_comment) {
	unlessnull (*_comment) { 
	    comment = gdupl (__FILE__, __LINE__, *_comment);
	}
    }
    for (;;) {
	while (isws(*p)) { ++p; }
	/* Wenn entweder das Zeilenende oder einer der Begrenzer erreicht ist,
	** dann sind wir fertig.
	*/
	if (*p == '\0' || strchr (delim, *p) != NULL) break;
	if (*p == '(') {
	    /* Wenn ein Kommentar gefunden wurde, dann mu dieser - an alle
	    ** bisherigen Kommentare angehngt werden - durch ein Leerzeichen
	    ** von diesen getrennt (Siehe: `apndfree').
	    */
	    /* Parsiere den Kommentar */
	    ifnull (sd = p_comment (&p, error)) FAILRET;
	    /* Hnge den neuen Kommentar an die shon gefundenen an. */
	    apndfree (&comment, sd, 1);
	} else if (*p == '.') {
	    /* Wenn ein `.' gefunden wurde, dann mu dieser an den `local part'
	    ** (oder aber den `domain part' angehngt werden - jedoch nur dann,
	    ** wenn nicht bereits ein Punkt vorher vorkam; zwei aufeinander
	    ** folgende Punkte sind laut RFC 822 nmlich nicht mglich, weil
	    ** ein Wort nicht leer sein kann.
	    */
	    if (dot) { *error = "empty word not allowed here"; FAILRET; }
	    ++p;
	    /* Da `apndfree' nur mit dynamisch erzeugten Zeichenketten arbeiten
	    ** kann, darf "." nicht einfach direkt bergeben werden; statt-
	    ** dessen wird eine dynamische Kopie von "." erzeugt, die dann als
	    ** zweites Argument an `apndfree' bergeben wird.
	    */
	    apndfree (&res, gdupl (__FILE__, __LINE__, "."), 0);
	    /* Mit dem gesetzten `dot'-Flag wird ausgedrckt, da nun kein
	    ** weiterer `.' folgen darf; stattdessen mu mindestens noch ein
	    ** Wort (oder eine Sub-Domain) folgen.
	    */
	    dot = 1;
	} else {
	    /* Behandle den Rest. Hier mu eine weitere Unterscheidung ge-
	    ** troffen werden. Wenn nmlich das letzte Zeichen kein `.' war,
	    ** dann kann dies bedeuten, da jetzt ein weiteres Wort einer
	    ** `phrase' folgt. Auf der anderen Seite ist das bei einer
	    ** Domain nicht mglich.
	    */
	    if (!dot) {
		if (dodot) { *error = "`.' required here"; FAILRET; }
		/* In jedem Fall jedoch ist die Parsierung in diesem Fall
		** beendet.
		*/
		break;
	    }
	    /* Wenn das letzte Zeichen ein `.' war, dann wird nun die als
	    ** Argument bergebene Funktion aufgerufen, um das nchste
	    ** Element des `locat part' (`domain part') zu parsieren.
	    */
	    ifnull (sd = pf (&p, error)) FAILRET;
	    /* Da wir hier keine `phrase' parsieren, sondern einen
	    ** `local part' (`domain part'), darf bei der Konkatenation
	    ** kein Leerzeichen eingefgt werden.
	    */
	    apndfree (&res, sd, 0);
	    dot = 0;
	}
    }
    /* Ein abschlieender `.' entspricht einem folgenden leeren Wort - was
    ** laut RFC 822 unzulssig ist.
    */
    if (dot) FAILRET;
    /* Weitere Spezialbehandlung fr Kommentare... */
    ifnull (_comment) {
	/* Es wurde keine Variable bergeben, die Kommentare aufnehmen soll.
	** Daher knnen alle bisher gesammelten Kommentare weggeworfen werden.
	*/
	gfree (comment);
    } else {
	/* Wenn ein Kommentar bergeben wurde, dann ist dieser oben kopiert
	** worden; da wir nun zu einem (erfolgreichen) Ende gekommen sind,
	** kann der ursprngliche Kommentar weggeworfen
	*/
	if (*_comment) {
	    gfree (*_comment);
	}
	/* und anschlieend durch den neuen ersetzt werden. */
	*_comment = comment;
    }
    /* Schreibe den neuen Wert des Zeichenzeigers zurck. */
    *_p = p;
    /* und gib den neu erzeugten `local part' (`domain part') zurck. */
    return res;
#   undef FAILRET
}

/* p_user (Parsierung des `local part') wird durch `p_ud' ausgedrckt. */

static char *p_user (char **_p, char **_comment, char **error) {
    return p_ud (_p, _comment, p_word, ",>;@", 0, error);
}

/* p_domain (Parsierung des `domain part') wird ebenfalls durch `p_ud'
** ausgedrckt.
*/
static char *p_domain (char **_p, char **_comment, char **error)
{
    return p_ud (_p, _comment, p_subdomain, ",>;:", 1, error);
}

/* p_phrase parsiert den Rest einer `phrase' (den Teil vor dem `:' bei
** einer Gruppen- bzw. `<' vor einer Routenadresse). Der erste Teil wurde
** bereits durch `p_user' bestimmt.
*/
static char *p_phrase (char **_p, char *pfx, char **_comment, char **error)
{
    char *p = *_p;
    char *comment = NULL;
    char *res = pfx, *sd;
    /* This macro should make the code of the function a bit mor readable.
    ** It ensures a consistent state before returning NULL in the case of
    ** an error.
    */
#   define FAILRET do { \
	*_p = p; \
	unlessnull (res) { gfree (res); } \
	unlessnull (comment) { gfree (comment); } \
	return NULL; \
    } while (0)

    /* The incoming value of `_comment' MUST be copied in order to leave it
    ** in an unchanged state when returning an error.
    */
    unlessnull (_comment) {
	unlessnull (*_comment) { 
	    comment = gdupl (__FILE__, __LINE__, *_comment);
	}
    }
    for (;;) {
	while (isws(*p)) { ++p; }
	/* Either the end of the (extended) line or one of the
	** characters `:' (begin of a group) or `<' begin of a route address
	** will terminate this function.
	*/
	if (*p == '\0' || *p == ':' || *p == '<') break;
	if (*p == '(') {
	    /* append the comment found to the comment string */
	    ifnull (sd = p_comment (&p, error)) FAILRET;
	    /* prepend the new comment part with a leading space - but only
	    ** if it's not the first part...
	    */
	    apndfree (&comment, sd, 1);
	} else if (*p == '.') {
	    /* append a dot to the phrase; a dot is concatenated immediately
	    ** (without a leading space).
	    */
	    apndfree (&res, gdupl (__FILE__, __LINE__, "."), 0);
	} else {
	    /* Any other ``word'' will be appended with a leading space */
	    ifnull (sd = p_word (&p, error)) FAILRET;
	    apndfree (&res, sd, 1);
	}
    }
    ifnull (_comment) {
	gfree (comment);
    } else {
	if (*_comment) { gfree (*_comment); }
	*_comment = comment;
    }
    *_p = p;
    return res;
#   undef FAILRET
}

/* `p_routeaddr' parsiert eine Routenadresse, d.h. eine Adresse im Format
**     <[@ host , ..., @ host :] user [@ host]>
** . Das Ergebnis ist ein Wert vom Typ `address_t'.
*/
static addrlist_t p_routeaddr (char **_p, char *host)
{
    char *p = *_p, *r, *error = NULL, *comment = NULL, *rs,
	 *user = NULL, *domain = NULL;
    route_t rl = (route_t) NULL, rc = (route_t) NULL;
    addrlist_t res;
    ++p; 
    for (;;) {
	/* berspringe Leerzeichen und Tabs */
	while (isws (*p)) { ++p; }
	/* Wenn das nun folgende Zeichen kein `@' ist, dann ist der `route'-Teil
	** zu ende.
	*/
	if (*p != '@') { break; }
	/* berspringe Leerzeichen und Tabs nach dem `@'. */
	++p; while (isws (*p)) { ++p; }
	/* Hole einen Domain-Namen. */
	ifnull (rs = p_domain (&p, &comment, &error)) { goto ERRRET; }
	/* Hnge den neuen Hostnamen vorn an die bisherige Route an. */
	ifnull (rl) {
	    rl = rc = (route_t) galloc (__FILE__, __LINE__,
					sizeof(struct route_s) +
					strlen (rs) + 1);
	} else {
	    rc->next = (route_t) galloc (__FILE__, __LINE__,
					 sizeof(struct route_s) +
					 strlen (rs) + 1);
	    rc = rc->next;
	}
	rc->next = (route_t) NULL;
	r = (char *) rc + sizeof(struct route_s);
	rc->route = r;
	strcpy (rc->route, rs);
	gfree (rs);
	/* Ein anderes Zeichen als `,' beendet beendet den `route'-Teil. */
	while (isws (*p)) { ++p; }
	if (*p != ',') {
	    /* Aber nur ein `:' ist GLTIG. */
	    if (*p != ':') {
		error = "invalid character in route address"; goto ERRRET;
	    }
	    ++p; break;
	}
	/* ein Komma ist ein gltiges Trennzeichen nach dem Host-Teil. */
	++p;
    }
    /* berspringe nochmal alle Leerzeichen und Tabs. */
    while (isws(*p)) { ++p; }
    /* Was jetzt folgen mu, ist ein gltiger `username' */
    ifnull (user = p_user (&p, &comment, &error)) { goto ERRRET; }
    if (*p == '>') { domain = gdupl (__FILE__, __LINE__, host); }
    if (*p != '@') { error = "invalid character in address"; goto ERRRET; }
    ++p; while (isws (*p)) { ++p; }
    ifnull (domain = p_user (&p, &comment, &error)) { goto ERRRET; }
    if (*p != '>') { error = "missing `>' in route address"; }
    ++p; while (isws (*p)) { ++p; }
    while (*p == '(') {
	char *comment1;
	ifnull (comment1 = p_comment (&p, &error)) { goto ERRRET; }
	apndfree (&comment, comment1, 1);
	while (isws (*p)) { ++p; }
    }
ERRRET:
    *_p = p;
    res = (addrlist_t) galloc (__FILE__, __LINE__, sizeof(struct addrlist_s));
    res->next = (addrlist_t) NULL;
    res->phrase = NULL;
    res->gphrase = NULL;
    res->comment = comment;
    res->user = user;
    res->domain = domain;
    res->route = rl;
    res->error = error;
    return res;
}

/* p_address wird rekursiv in `p_group' aufgerufen; daher mu hier eine
** Vorwrtsdeklaration erfolgen.
*/
static addrlist_t p_address (char **_p, char *host, int nogroup);

/* Parsiere eine Gruppenadresse, d.h. eine Adresse von der Form
** `: addr, ... ;'
*/
static addrlist_t p_group (char **_p, char *host)
{
    addrlist_t lh = (addrlist_t) NULL, lc = (addrlist_t) NULL, addr;
    char *p = *_p;
    /* Hmmm... - eine Gruppenadresse besteht aus einem `:', gefolgt von lauter
    ** einzelnen - durch `,' getrennten - Adressen, gefolgt von einem `;'.
    */
    ++p;
    for (;;) {
	addr = p_address (&p, host, 1);
	ifnull (lh) {
	    lh = lc = addr;
	} else {
	    lc->next = addr;
	    lc = lc->next;
	}
	lc->next = (addrlist_t) NULL;
	if (*p == ';') { ++p; break; }
	if (*p != ',') {
	    /* Invalid member address in group... */
	    lc->error = "invalid char in group address";
	    break;
	}
	++p;
    }
    *_p = p;
    return lh;
}

/* Parsiere eine E-Mail-Adresse; abhngig von dem dritten Argument `nogroup'
** wird dabei eine Gruppenadresse zugelassen (nogroup = 0) oder nicht
** (nogroup = 1).
*/
static addrlist_t p_address (char **_p, char *host, int nogroup)
{
    char *p = *_p;
    char *phrase = NULL, *comment = NULL, *user = NULL, *domain = NULL,
	 *error = NULL;
    addrlist_t res;
#   define pa_INIT do { \
	unlessnull (phrase) { gfree (phrase); } \
	unlessnull (comment) { gfree (comment); } \
	unlessnull (user) { gfree (user); } \
	unlessnull (domain) { gfree (domain); } \
	*_p = p; \
    } while (0)

    /* berspringe Leerzeichen und Tabs. */
    while (isws (*p)) { ++p; }

    /* Eine nicht existierende Adresse ist ein Fehler! */
    if (!*p) { error = "missing address"; goto ERRRET; }

    /* Wenn ein `:' gefunden wurde, dann folgt eine `group address', d.h.
    ** eine Adresse im Format `: address, ..., address ;'.
    */
    if (*p == ':') {
	/* Diese `group address' ist aber nur dann gltig, wenn `nogroup' nicht
	** gesetzt ist. Denn - um diese Routine nicht vollstndig neu schreiben
	** zu mssen, um Adressen innerhalb einer Gruppe zu parsieren, bergebe
	** ich dieses Flag als Argument. (Eine Gruppe innerhalb einer Gruppe ist
	** laut RFC 822 nicht zulssig.)
	*/
	if (nogroup) { error = "group address not allowed here"; goto ERRRET; }
	/* Ansonsten wird eine Gruppenadresse parsiert und zurckgegeben. */
	*_p = p; return p_group (_p, host);
    }

    /* Wenn ein `<' gefunden wurde, dann folgt eine `route address',
    ** d.h. eine Adresse im Format `<[@host, ..., @host :] user@domain>'.
    */
    if (*p == '<') { *_p = p; return p_routeaddr (_p, host); }

    /* Was nun folgt, kann sowohl eine einfache Mail-Adresse als auch
    ** eine `group address' oder eine `route address' sein (beides dann
    ** mit einleitender `phrase'.
    */
    ifnull (user = p_user (&p, &comment, &error)) { goto ERRRET; }

    /* Wenn das nchste Zeichen nach dem parsierten `username' ein `,' ist,
    ** oder wenn das Zeilenende erreicht ist, dann haben wir eine einfache
    ** Adresse gefunden, die nur aus dem `username' besteht. In diesem Fall
    ** mu der als zweites Argument bergebene `host' als Domainname
    ** eingesetzt werden.
    */
    if (!*p || *p == ',') {
	/* Das ist hier keine Fehlersituation; der Punkt `ERRRET' ist
	** lediglich der normale Aussprungpunkt aus der Routine.
	*/
	domain = gdupl (__FILE__, __LINE__, host); goto ERRRET;
    }

    /* Wenn das nchste Zeichen ein `:' ist, dann handelt es sich um eine
    ** Gruppenadresse. In diesem Fall mu die Gruppe parsiert werden - jedoch
    ** nur, wenn nogroup nicht gesetzt ist.
    */
    if (*p == ':') {
	if (nogroup) { error = "group address not allowed here"; goto ERRRET; }
	*_p = p; res = p_group (_p, host);
	unlessnull (res) {
	    res->gphrase = user; user = NULL;
	    apndfree (&comment, res->comment, 1);
	    res->comment = comment;
	}
	/* Da in diesem Fall die Funktion `p_group' bereits die vollstndige
	** `addrlist'-Struntur geliefert hat, verlassen wir `p_address'
	** vorzeitig.
	*/
	return res;
    }
    /* Wenn das nchste Zeichen ein `<' ist, dann handelt es sich um eine
    ** `route address' plus einleitender `phrase'.
    */
    if (*p == '<') {
	*_p = p; res = p_routeaddr (_p, host);
	unlessnull (res) {
	    apndfree (&user, res->phrase, 1); res->phrase = user; user = NULL;
	    apndfree (&comment, res->comment, 1); res->comment = comment;
	    res->gphrase = NULL;
	}
	return res;
    }
    /* Wenn das folgende Zeichen ein `@' ist, dann handelt es sich um eine
    ** einfache E-Mail-Adresse.
    */
    if (*p == '@') {
	++p;
	domain = p_domain (&p, &comment, &error);
	goto ERRRET;
    }
    /* Parsiere den Rest der einleitenden `phrase'. An dieser Stelle besteht
    ** folgender Zustand:
    ** i)   Eine Adresse (evtl. ein einzelnes Wort) wurde erfolgreich
    **      parsiert;
    ** ii)  Es wurde kein `@' danach gefunden.
    ** Das einzige, was nun noch folgen darf ist entweder eine Gruppen-
    ** oder eine Routenadresse - nach dem Rest der `phrase'.
    ** Bestimmen wir also zunchst den Rest der `phrase'.
    */
    phrase = user; user = NULL;
    ifnull (phrase = p_phrase (&p, phrase, &comment, &error)) { goto ERRRET; }
    /* Wenn jetzt noch ein `:' folgt, dann handelt es sich um eine
    ** Gruppenadresse.
    */
    if (*p == ':') {
	/* Die darf nur dann parsiert werden, wenn `nogroup' nicht gesetzt
	** ist.
	*/
	if (nogroup) { error = "group address not allowed here"; goto ERRRET; }
	*_p = p; res = p_group (_p, host);
	unlessnull (res) {
	    res->gphrase = user; user = NULL;
	    apndfree (&comment, res->comment, 1); res->comment = comment;
	}
	/* Wieder ein vorzeitiger Aussprung - weil das Ergebnis nun bereits
	** vollstndig vorliegt.
	*/
	return res;
    }

    /* Wenn nach der `phrase' ein `<' folgt, dann handelt es sich um eine
    ** `route address'.
    */
    if (*p == '<') {
	*_p = p; res = p_routeaddr (_p, host);
	unlessnull (res) {
	    apndfree (&phrase, res->phrase, 1); res->phrase = phrase;
	    apndfree (&comment, res->comment, 1); res->comment = comment;
	    res->gphrase = NULL;
	}
	return res;
    }
    /* Alles andere ist ungltig. */
    error = "invalid address";
ERRRET:
    /* Schreibe zunchst den (mglicherweise genderten) Zeichenzeiger
    ** zurck.
    */
    *_p = p;
    /* Nun mu noch das Ergenis erzeugt werden. */
    res = (addrlist_t) galloc (__FILE__, __LINE__, sizeof(struct addrlist_s));
    /* Dabei werden alle oben gesammelten Daten eingesammelt. */
    res->next = (addrlist_t) NULL;
    res->gphrase = NULL;
    res->phrase = phrase;
    res->comment = comment;
    res->user = user;
    res->domain = domain;
    res->route = (route_t) NULL;
    res->error = error;
    return res;
#   undef pa_INIT
}

/******************************************************************
**                                                               **
** Hier folgen die Funktionen, die eigentlich exportiert werden. **
**                                                               **
******************************************************************/

/* rfc822_addrlist Parsiert smtliche vorkommenden Adressen und
** gibt eine Liste mit den gefundenen Adressen zurck.
**/
addrlist_t rfc822_addrlist (char **_p, char *host)
{
    addrlist_t al = (addrlist_t) NULL, ac = (addrlist_t) NULL;
    for (;;) {
	ifnull (al) {
	    al = ac = p_address (_p, host, 0);
	} else {
	    ac->next = p_address (_p, host, 0);
	    while (noNULL(ac->next)) { ac = ac->next; }
	}
	if (**_p != ',') {
	    if (**_p) { al->error = "`,' expected"; }
	    break;
	}
	++(*_p);
    }
    return al;
}

/* rfc822_address parsiert eine einzelne Adresse; da diese jedoch auch eine
** Gruppenadresse sein kann, knnen hier auch mehrere Adressen im Ergebnis
** stehen.
*/
addrlist_t rfc822_address (char **_p, char *host)
{
    return p_address (_p, host, 0);
}

/* rfc822_singleaddr parsiert genau eine Adresse, wobei Gruppenadressen
** ausdrcklich ausgeschlossen werden. Darum wird auch nur genau eine
** Adresse zurckgegeben.
*/
addrlist_t rfc822_singleaddr (char **_p, char *host)
{
    return p_address (_p, host, 1);
}

void address_free (addrlist_t al);

static char *single2str (addrlist_t al, int full);

/* rfc822_addrstrg parsiert eine Routingadresse gem RFC 822 oder eine
** normale E-Mail-Adresse (keine Gruppenadresse) und gibt das Ergebnis
** in Form einer Routingadresse zurck.
*/
char *rfc822_addrstrg (char **_p, char *host)
{
    addrlist_t addr;
    char *res = NULL;
    addr = p_address (_p, host, 1);
    ifnull (addr) { return NULL; }
    if (addr->error) {
	fprintf (stderr, "Error: %s\n", addr->error);
	address_free (addr); return NULL;
    }
    res = single2str (addr, 0);
    address_free (addr);
    return res;
}

/* route_free gibt den gesamten Speicherplatz eine Route frei. */
void route_free (route_t rt)
{
    route_t nx;
    while (noNULL(rt)) {
	nx = rt->next;
	gfree (rt);
	rt = nx;
    }
}

/* address_free gibt die gesamte von `p_address' erzeugte Adressliste
** (einschl. der Routenlisten) frei.
*/
void address_free (addrlist_t al)
{
    addrlist_t nx;
    while (noNULL(al)) {
	nx = al->next;
	unlessnull (al->gphrase) { gfree (al->gphrase); }
	unlessnull (al->phrase) { gfree (al->phrase); }
	unlessnull (al->comment) { gfree (al->comment); }
	unlessnull (al->user) { gfree (al->user); }
	unlessnull (al->domain) { gfree (al->domain); }
	route_free (al->route);
	al = nx;
    }
}

/* `single2str' gibt die angegebene einzelne Adresse als Zeichenkette
** zurck. Wenn `full == 0' ist, dann wird die reine Routenadresse
** zurckgegeben; ist hingegen `full == 1', dann werden auch eine fhrende
** `phrase' und ein enthaltener Kommentar' zurckgegeben.
*/
static char *single2str (addrlist_t al, int full)
{
    char *res = NULL, *p;
    route_t r, rt = al->route;
    size_t ml = 0;
    for (r = rt; noNULL(r); r = r->next) {
	ml += strlen (r->route) + 3;
	    /* `,' + ` ' + `@' + route */
    }
    /* two bytes too much for the first route (`@' + route), but there is an
    ** additional `:' at the end of the route list - so it's only one byte
    ** too much.
    */
    ml += 2; /* `<' + `>' */
    if (full) {
	if (al->phrase) { ml += strlen (al->phrase) + 1; } /* phrase + ` ' */
	if (al->comment) { ml += strlen (al->comment) + 3; }
	    /* ` ' + `(' + comment + ')' */
    }
    ml += strlen (al->user) + strlen (al->domain) + 2;
	/* user + `@' + domain + `\0' */
    /* Ab hier gehe ich von einer nicht-leeren Route aus. */
    res = galloc (__FILE__, __LINE__, ml);
    p = res;
    if (full && noNULL(al->phrase)) {
	ml = strlen (al->phrase); memcpy (p, al->phrase, ml); p += ml;
	*p++ = ' ';
	/* res == "phrase " */
    }
    *p++ = '<'; /* res == "...<" */
    unlessnull (rt) {
	*p++ = '@'; /* res == "...<@" */
	ml = strlen (rt->route); memcpy (p, rt->route, ml); p += ml;
	    /* res == "...<@route1" */
	for (r = rt->next; noNULL(r); r = r->next) {
	    *p++ = ','; *p++ = ' '; *p++ = '@';
		/* res == "...<@route1..., @" */
	    ml = strlen (r->route); memcpy (p, r->route, ml); p += ml;
		/* res == "...<@route1..., @routeN" */
	}
	*p++ = ':'; *p++ = ' ';
	/* res == "...<@route1..., @routeN: " */
    }
    ml = strlen (al->user); memcpy (p, al->user, ml); p += ml;
    *p++ = '@'; /* res == "...<...user@" */
    ml = strlen (al->domain); memcpy (p, al->domain, ml); p += ml;
    *p++ = '>'; /* res == "...<...user@domain>" */
    if (full && noNULL(al->comment)) {
	*p++ = ' '; *p++ = '('; /* res == "...<...user@domain> (" */
	ml = strlen (al->comment); memcpy (p, al->comment, ml); p += ml;
	*p++ = ')'; /* res == "...<...user@domain> (comment)" */
    }
    *p = '\0'; /* res == "...<...user@domain>..." */
    return res;
}

char *rfc822_addr2strg (addrlist_t addr, int full)
{
    /*addrlist_t pa;*/
    return NULL;
}
